==> TundereBird-0.23.tar.bz2: TundereBird-0.23/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.WebClients/ImapAppendMessageBodyStream.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Threading;

namespace Smdn.Net.Imap4.WebClients {
  internal class ImapAppendMessageBodyStream : Stream {
    public override bool CanSeek {
      get { RejectDisposed(); return false; }
    }

    public override bool CanRead {
      get { RejectDisposed(); return true; }
    }

    public override bool CanWrite {
      get { RejectDisposed(); return true; }
    }

    public override bool CanTimeout {
      get { RejectDisposed(); return true; }
    }

    public override long Position {
      get { RejectDisposed(); return writeOffset; }
      set { RejectDisposed(); writeOffset = value; }
    }

    public override long Length {
      get
      {
        RejectDisposed();
        lengthSetEvent.WaitOne();
        return length;
      }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return writeTimeout; }
      set
      {
        RejectDisposed();
        if (value < -1)
          throw new ArgumentOutOfRangeException("value", "must be greater than or equals to -1");
        writeTimeout = value;
      }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return readTimeout; }
      set
      {
        RejectDisposed();
        if (value < -1)
          throw new ArgumentOutOfRangeException("value", "must be greater than or equals to -1");
        readTimeout = value;
      }
    }

    public ImapAppendMessageBodyStream(int readWriteTimeout)
    {
      if (readWriteTimeout < -1)
        throw new ArgumentOutOfRangeException("readWriteTimeout", "must be greater than or equals to -1");

      readTimeout   = readWriteTimeout;
      writeTimeout  = readWriteTimeout;

      buffer = new MemoryStream(4096);
    }

    public override void Close()
    {
      UpdateLength();

      base.Close();
    }

    internal void UpdateLength()
    {
      if (!lengthSetEvent.WaitOne(0)) {
        this.length = buffer.Length;

        lengthSetEvent.Set();
      }
    }

    internal void InternalDispose()
    {
      if (disposed)
        return;

      Dispose();

      lengthSetEvent.Close();
      lengthSetEvent = null;

      writtenEvent.Close();
      writtenEvent = null;

      buffer.Close();
      buffer = null;

      disposed = true;
    }

    public override void SetLength(long @value)
    {
      if (lengthSetEvent.WaitOne(0))
        throw new InvalidOperationException("length has been set");

      length = @value;
      lengthSetEvent.Set();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    public override void Flush()
    {
      // do nothing
    }

    private byte[] singleByteBuffer = new byte[1];

    public override void WriteByte(byte @value)
    {
      singleByteBuffer[0] = @value;

      Write(singleByteBuffer, 0, 1);
    }

    public override void Write(byte[] src, int offset, int count)
    {
      RejectDisposed();

      if (!Monitor.TryEnter(readWriteLock, writeTimeout))
        throw new TimeoutException("write timeout");

      try {
        writeOffset = buffer.Seek(writeOffset, SeekOrigin.Begin);

        buffer.Write(src, offset, count);

        writeOffset += count;

        writtenEvent.Set();
      }
      finally {
        Monitor.Exit(readWriteLock);
      }

      WaitForReading();
    }

    private void WaitForReading()
    {
      if (lengthSetEvent.WaitOne(0) && 40960 < writeOffset - readOffset)
        Thread.Sleep(50);
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      lengthSetEvent.WaitOne();

      if (!Monitor.TryEnter(readWriteLock, readTimeout))
        throw new TimeoutException("read timeout");

      try {
        readOffset = buffer.Seek(readOffset, SeekOrigin.Begin);

        var bytesToRead = (readOffset + count < length)
          ? count
          : (int)(length - readOffset);

      RetryRead:
        var read = buffer.Read(dest, offset, bytesToRead);

        if (read == 0 && readOffset < length) {
          if (writtenEvent.WaitOne(readTimeout))
            goto RetryRead;
          else
            throw new TimeoutException("buffer underrun");
        }

        readOffset += read;

        return read;
      }
      finally {
        Monitor.Exit(readWriteLock);
      }
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private long writeOffset  = 0L;
    private long readOffset   = 0L;
    private long length       = 0L;
    private int writeTimeout  = Timeout.Infinite;
    private int readTimeout   = Timeout.Infinite;

    private MemoryStream buffer;
    private ManualResetEvent lengthSetEvent = new ManualResetEvent(false);
    private AutoResetEvent writtenEvent = new AutoResetEvent(false);
    private object readWriteLock = new object();
  }
}

==> TundereBird-0.24.tar.bz2: TundereBird-0.24/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.WebClients/ImapAppendMessageBodyStream.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Threading;

namespace Smdn.Net.Imap4.WebClients {
  internal class ImapAppendMessageBodyStream : Stream {
    public override bool CanSeek {
      get { RejectDisposed(); return false; }
    }

    public override bool CanRead {
      get { RejectDisposed(); return true; }
    }

    public override bool CanWrite {
      get { RejectDisposed(); return true; }
    }

    public override bool CanTimeout {
      get { RejectDisposed(); return true; }
    }

    public override long Position {
      get { RejectDisposed(); return writeOffset; }
      set { RejectDisposed(); writeOffset = value; }
    }

    public override long Length {
      get
      {
        RejectDisposed();
        lengthSetEvent.WaitOne();
        return length;
      }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return writeTimeout; }
      set
      {
        RejectDisposed();
        if (value < -1)
          throw new ArgumentOutOfRangeException("value", "must be greater than or equals to -1");
        writeTimeout = value;
      }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return readTimeout; }
      set
      {
        RejectDisposed();
        if (value < -1)
          throw new ArgumentOutOfRangeException("value", "must be greater than or equals to -1");
        readTimeout = value;
      }
    }

    public ImapAppendMessageBodyStream(int readWriteTimeout)
    {
      if (readWriteTimeout < -1)
        throw new ArgumentOutOfRangeException("readWriteTimeout", "must be greater than or equals to -1");

      readTimeout   = readWriteTimeout;
      writeTimeout  = readWriteTimeout;

      buffer = new MemoryStream(4096);
    }

    public override void Close()
    {
      UpdateLength();

      base.Close();
    }

    internal void UpdateLength()
    {
      if (!lengthSetEvent.WaitOne(0)) {
        this.length = buffer.Length;

        lengthSetEvent.Set();
      }
    }

    internal void InternalDispose()
    {
      if (disposed)
        return;

      Dispose();

      lengthSetEvent.Close();
      lengthSetEvent = null;

      writtenEvent.Close();
      writtenEvent = null;

      buffer.Close();
      buffer = null;

      disposed = true;
    }

    public override void SetLength(long @value)
    {
      if (lengthSetEvent.WaitOne(0))
        throw new InvalidOperationException("length has been set");

      length = @value;
      lengthSetEvent.Set();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    public override void Flush()
    {
      // do nothing
    }

    private byte[] singleByteBuffer = new byte[1];

    public override void WriteByte(byte @value)
    {
      singleByteBuffer[0] = @value;

      Write(singleByteBuffer, 0, 1);
    }

    public override void Write(byte[] src, int offset, int count)
    {
      RejectDisposed();

      if (!Monitor.TryEnter(readWriteLock, writeTimeout))
        throw new TimeoutException("write timeout");

      try {
        writeOffset = buffer.Seek(writeOffset, SeekOrigin.Begin);

        buffer.Write(src, offset, count);

        writeOffset += count;

        writtenEvent.Set();
      }
      finally {
        Monitor.Exit(readWriteLock);
      }

      WaitForReading();
    }

    private void WaitForReading()
    {
      if (lengthSetEvent.WaitOne(0) && 40960 < writeOffset - readOffset)
        Thread.Sleep(50);
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      lengthSetEvent.WaitOne();

      if (readOffset == length)
        return 0;

      var ret = 0;

      for (;;) {
        if (readOffset == writeOffset && !writtenEvent.WaitOne(readTimeout))
          throw new TimeoutException("buffer underrun");

        if (!Monitor.TryEnter(readWriteLock, readTimeout))
          throw new TimeoutException("read timeout");

        try {
          readOffset = buffer.Seek(readOffset, SeekOrigin.Begin);

          var read = buffer.Read(dest,
                                 offset,
                                 (readOffset + count < length)
                                   ? count
                                   : (int)(length - readOffset));

          readOffset  += read;
          count       -= read;
          ret         += read;
        }
        finally {
          Monitor.Exit(readWriteLock);
        }

        if (0 < count && readOffset < length)
          continue;

        return ret;
      }
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private long writeOffset  = 0L;
    private long readOffset   = 0L;
    private long length       = 0L;
    private int writeTimeout  = Timeout.Infinite;
    private int readTimeout   = Timeout.Infinite;

    private MemoryStream buffer;
    private ManualResetEvent lengthSetEvent = new ManualResetEvent(false);
    private AutoResetEvent writtenEvent = new AutoResetEvent(false);
    private object readWriteLock = new object();
  }
}
